#!/bin/bash


usage() {
    echo "usage: ${0##*/} list" >&2
    echo "usage: ${0##*/} [-n NAME] [-v VERSION] [-m MISSING_DEP] rpmstuff" >&2
    exit 2
}

die() {
    echo "$1" >&2
    exit 3
}

possible_cases() {
    #
    # All "mathematically" possible flag combinations
    #
    # from rpm(1):
    #
    # > The  format  of  the  output  is a string of 9 characters, a possible
    # > attribute marker:
    # >
    # > c %config configuration file.
    # > d %doc documentation file.
    # > g %ghost file (i.e. the file contents are not included in the package p ayload).
    # > l %license license file.
    # > r %readme readme file.
    # >
    # > from the package header, followed by the file name.  Each  of  the  9
    # > characters  denotes the result of a comparison of attribute(s) of the
    # > file to the value of those attribute(s) recorded in the database.   A
    # > single  "." (period) means the test passed, while a single "?" (ques‐
    # > tion mark) indicates the test could not be performed (e.g. file  per‐
    # > missions  prevent  reading). Otherwise, the (mnemonically emBoldened)
    # > character denotes failure of the corresponding --verify test:
    # >
    # > S file Size differs
    # > M Mode differs (includes permissions and file type)
    # > 5 digest (formerly MD5 sum) differs
    # > D Device major/minor number mismatch
    # > L readLink(2) path mismatch
    # > U User ownership differs
    # > G Group ownership differs
    # > T mTime differs
    # > P caPabilities differ
    #
    local ftype                 # rpm-packable file type
    local frole                  # RPM attributr maker (%config, %dir..., N for none)
    local s m d h l u g t p     # chars (0 for ., 1 for flag, u for '?')
    for ftype in dev fil lnk dir;
    do
        for frole in c C d g l r N;
        do
            for s in 0 1 u; do
            for m in 0 1 u; do
            for h in 0 1 u; do
            for d in 0 1 u; do
            for l in 0 1 u; do
            for u in 0 1 u; do
            for g in 0 1 u; do
            for t in 0 1 u; do
            for p in 0 1 u; do
                echo $ftype.r$frole.S$s.M$m.5$h.D$d.L$l.U$u.G$g.T$t.P$p;
            done
            done
            done
            done
            done
            done
            done
            done
            done
            echo $ftype.r$frole.missing
        done
    done
}

valid_cases() {
    #
    # All interesting (valid *and* somewhat probable) cases
    #
    # Since number of all possible cases is threatening (thousands),
    # we're weeding out a lot here.  We're pretending that:
    #
    #  *  nobody uses %readme directive,
    #
    #  *  nobody uses %license directive (OK, well, it *is* on Fedora
    #     wiki as an example, but 6.8 rpmbuild expands it as macro
    #     so we are either broken or it's not really supported),
    #
    #  *  dir's hash and size are ignored by RPM (or we don't know what
    #     it means)
    #
    #  *  dirs cannot have link target,
    #
    #  *  dirs and links cannot have hash,
    #
    #  *  dirs cannot have %-directive,
    #
    #  *  size cannot change without hash also changing,
    #
    #  *  nobody packages device files these days,
    #
    #  *  all rpm -V tests can be done as we're running as root,
    #
    #  *  links and dirs cannot have capabilities(7).
    #
    local o='[01]'      # both states
    local x=0           # only "passed" (attr. unchanged)
    grep -x \
      -e "dir.rN.S$x.M$o.5$x.D$x.L$x.U$o.G$o.T$o.P$x" \
      -e "fil.r..S$o.M$o.5$o.D$x.L$x.U$o.G$o.T$o.P$o" \
      -e "lnk.r..S$o.M$o.5$x.D$x.L$o.U$o.G$o.T$o.P$x" \
      -e "....r..missing" \
      | grep -vx \
      -e "dev.*" \
      -e "dir.r[^N].*" \
      -e "....rr.*" \
      -e "....rl.*" \
      | grep \
      -e "S0.M..50" \
      -e "S0.M..51" \
      -e "S1.M..51" \
      -e "missing"
}

make_mauler() {
    #
    # Print body of a mauler script that should be added to rpm
    #
    #shellcheck disable=SC2016
    {
        echo '#!/bin/bash'
        echo ''
        echo 'warn() {'
        echo '    echo "$1" >&2'
        echo '}'
        echo ''
        echo 'debug() {'
        echo '    $TORTURED_DEBUG || return 0'
        echo '    warn "$1"'
        echo '}'
        echo ''
        echo 'die() {'
        echo '    warn "$1"'
        echo '    exit 3'
        echo '}'
        echo ''
        echo 'TORTURED_DEBUG=${TORTURED_DEBUG:-false}'
        echo 'fname=none'
        echo 'ftype=none    # dev fil lnk dir'
        echo 'frole=none    # c d g l r N'
        echo 's=none        # 0 1'
        echo 'm=none        # ″'
        echo 'h=none        # ″'
        echo 'd=none        # ″'
        echo 'l=none        # ″'
        echo 'u=none        # ″'
        echo 'g=none        # ″'
        echo 't=none        # ″'
        echo 'p=none        # ″'
        echo 'warn TORTURED_DEBUG=$TORTURED_DEBUG'
        echo ''
        echo "rpm -ql $Name \\"
        echo "  | grep -vx /usr/bin/$Name-maul \\"
        echo "  | grep -vx /$Name \\"
        echo '  | while read path;'
        echo '    do'
        echo ''
        echo '        # remove items to be missing (no mauling possible)'
        echo '        #'
        echo '        case $path in'
        echo '            *.missing) rm -rf "$path"; continue ;;'
        echo '        esac'
        echo ''
        echo '        # create %ghost files (no mauling possible)'
        echo '        #'
        echo '        case $path in'
        echo '            lnk.rg.*) ln -s /dev/null/untracked/link/location "$path"; continue ;;'
        echo '            *.rg.*)   echo "untracked content" > "$path"; continue ;;'
        echo '        esac'
        echo ''
        echo '        # parse attributes from filename:'
        echo '        #                                   # path=/uglyrpm/fil.rg.S0.M0.50.D0.L0.U0.G0.T0.P0'
        echo '        fname=${path##*/}                   # fname=fil.rg.S0.M0.50.D0.L0.U0.G0.T0.P0'
        echo '        ftype=${fname:0:3}                  # ftype=fil'
        echo '        frole=${fname:5:1}                  # frole=g'
        echo '        s=${fname:8:1}                      # s=0'
        echo '        m=${fname:11:1}                     # m=0'
        echo '        h=${fname:14:1}                     # h=0'
        echo '        d=${fname:17:1}                     # d=0'
        echo '        l=${fname:20:1}                     # l=0'
        echo '        u=${fname:23:1}                     # u=0'
        echo '        g=${fname:26:1}                     # g=0'
        echo '        t=${fname:29:1}                     # t=0'
        echo '        p=${fname:32:1}                     # p=0'
        echo '        debug "fname=$fname"'
        echo '        debug "frole=$frole"'
        echo '        debug "s=$s m=$m h=$h d=$d l=$l u=$u g=$g t=$t p=$p"'
        echo ''
        echo '        # do the "mauling action" for specific attribute, if needed'
        echo '        #'
        echo '        case $ftype:$s in'
        echo '            *:0)   :                                                ;;'
        echo '            fil:1) echo padding >>"$path"                           ;;'
        echo '            dir:1) touch >>"$path/unexpected_file"                  ;;'
        echo '            *:1)   warn "unsupported: maul S of $ftype"             ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, s=$s in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$m in'
        echo '            *:0)   :                                                ;;'
        echo '            *:1)   chmod g-r "$path"                                ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, m=$m in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$h in'
        echo '            *:0)   :                                                ;;'
        echo '            fil:1) sed -i -e s/\\./X/ "$path"                       ;;'
        echo '            *:1)   warn "unsupported: maul H of $ftype"             ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, h=$h in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$d in'
        echo '            *:0)   :                                                ;;'
        echo '            *:1)   warn "unsupported: maul D of $ftype"             ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, d=$d in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$l in'
        echo '            *:0)   :                                                ;;'
        echo '            lnk:1) rm "$path"; ln -s /dev/null "$path"              ;;'
        echo '            *:1)   warn "unsupported: maul L of $ftype"             ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, l=$l in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$u in'
        echo '            *:0)   :                                                ;;'
        echo '            *:1)   chown -h daemon "$path"                          ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, u=$u in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$g in'
        echo '            *:0)   :                                                ;;'
        echo '            *:1)   chown -h :daemon "$path"                         ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, g=$g in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$t in'
        echo '            *:0)   :                                                ;;'
        echo '            *:1)   touch -h -d "next Friday" "$path"                ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, t=$t in $path" ;;'
        echo '        esac'
        echo '        case $ftype:$p in'
        echo '            *:0)   :                                                ;;'
        echo '            *:1)   setcap cap_kill+ep "$path"                       ;;'
        echo '            *:)    die "parsing error: ftype=$ftype, p=$p in $path" ;;'
        echo '        esac'
        echo '    done'
        test -n "$MissingDep" && {
            echo ''
            echo "rpm --nodeps -e '$MissingDep'"
        }
    }
}

print_spec() {
    #
    # Print specfile body
    #

    local app_root="/$Name"
    local buildroot="%_buildrootdir/$Name-$Version-1.el6.x86_64"
    local buildsrc="%_builddir/$Name-$Version"

    echo "Name:		$Name"
    echo "Version:	$Version"
    echo "Release:	1%{?dist}"
    echo "Summary:	$Name - a very ugly test RPM"
    echo "License:	LGPLv2"
    echo "Source0:	%{name}-%{version}.tar.gz"
    echo "BuildArch:	noarch"
    test -n "$MissingDep" && echo "Requires:	$MissingDep"
    echo "%description"
    echo "a very ugly test RPM"
    echo "%prep"
    echo "%setup -q"
    echo "%build"

    echo "%postun"
    echo "rm -rf /$Name"

    echo "%install"
    echo pwd
    echo "mkdir -p '$buildroot/$app_root'"
    echo "mkdir -p '$buildroot/usr/bin'"
    echo "install -m 0755 '$buildsrc/$Name-maul' $buildroot/usr/bin"
    "$0" list \
      | while read tc;
        do
            case $tc in
                dir.*) echo "mkdir -p '$buildroot/$app_root/$tc'"         ;;
                fil.*) echo "echo '$tc' > '$buildroot/$app_root/$tc'"     ;;
                lnk.*) echo "ln -s /etc/fstab '$buildroot/$app_root/$tc'" ;;
            esac
        done
    echo ""

    echo "%files"
    echo "/usr/bin/$Name-maul"
    echo "%dir /$Name"
    "$0" list \
      | while read tc;
        do
            case $tc in
                dir.*)  echo "%dir     $app_root/$tc" ;;
                *.rN.*) echo          "$app_root/$tc" ;;
                *.rc.*) echo "%config  $app_root/$tc" ;;
                *.rC.*) echo "%config(noreplace) $app_root/$tc" ;;
                *.rd.*) echo "%doc     $app_root/$tc" ;;
                *.rg.*) echo "%ghost   $app_root/$tc" ;;
                *.rl.*) echo "%license $app_root/$tc" ;;
                *.rr.*) echo "%readme  $app_root/$tc" ;;
            esac
        done

}

make_tarball() {
    local dirname="$Name-$Version"
    mkdir "$dirname"
    echo "this is a pseudo sources tarball for a testing RPM" > "$dirname/README"
    make_mauler > "$dirname/$Name-maul"
    tar -czf "$dirname.tar.gz" "$dirname"
}

main() {
    local cmd
    local Name
    local Version
    local MissingDep
    Name="uglyrpm"
    Version="0.0.0"
    while true; do case $1 in
        -m) MissingDep="$2"; shift 2 || usage ;;
        -n) Name="$2";    shift 2 || usage ;;
        -v) Version="$2"; shift 2 || usage ;;
        -*) usage ;;
        *)  break ;;
    esac done
    cmd="$1"
    test -n "$Name"             || die "NAME must not be empty"
    grep -q "[^a-z]" <<<"$Name" && die "NAME must be alphabetic: $Name"
    case $cmd in
        list)
            possible_cases | valid_cases
            ;;
        rpmstuff)
            print_spec > "$Name.spec"
            make_tarball
            ;;
        *)
            usage
            ;;
    esac
}

main "$@"
